/*
 *  ICARUS -  Interactive platform for Corpus Analysis and Research tools, University of Stuttgart
 *  Copyright (C) 2012-2013 Markus Gärtner and Gregor Thiele
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses.
 *
 * $Revision: 389 $
 * $Date: 2015-04-23 12:19:15 +0200 (Do, 23 Apr 2015) $
 * $URL: https://subversion.assembla.com/svn/icarusplatform/trunk/Icarus/plugins/de.ims.icarus.plugins.core/source/de/ims/icarus/plugins/core/ShutdownDialog.java $
 *
 * $LastChangedDate: 2015-04-23 12:19:15 +0200 (Do, 23 Apr 2015) $
 * $LastChangedRevision: 389 $
 * $LastChangedBy: mcgaerty $
 */
package de.ims.icarus.plugins.core;

import java.awt.BorderLayout;
import java.awt.Dialog.ModalityType;
import java.awt.Dimension;
import java.util.List;
import java.util.logging.Level;

import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.border.EmptyBorder;

import de.ims.icarus.Core;
import de.ims.icarus.Core.NamedRunnable;
import de.ims.icarus.logging.LoggerFactory;

/**
 * @author Markus Gärtner
 * @version $Id: ShutdownDialog.java 389 2015-04-23 10:19:15Z mcgaerty $
 *
 */
public final class ShutdownDialog {

	private JProgressBar progressBar;
	private JDialog dialog;
	private JLabel label;

	private static volatile ShutdownDialog instance;

	private boolean shutdownStarted = false;

	public static ShutdownDialog getDialog() {
		if(instance==null) {
			synchronized (ShutdownDialog.class) {
				if(instance==null) {
					instance = new ShutdownDialog();
				}
			}
		}

		return instance;
	}

	private ShutdownDialog() {
		// no-op
	}

	public synchronized void shutdown() {

		if(!SwingUtilities.isEventDispatchThread()) {
			SwingUtilities.invokeLater(new Runnable() {

				@Override
				public void run() {
					shutdown();
				}
			});
			return;
		}

		if(shutdownStarted) {
			return;
		}
		shutdownStarted = true;

		// Build panel

		dialog = new JDialog();
		dialog.setIconImages(Core.getIconImages());
		dialog.setModalityType(ModalityType.APPLICATION_MODAL);

		progressBar = new JProgressBar();
		progressBar.setStringPainted(false);
		progressBar.setBorder(new EmptyBorder(5, 15, 10, 15));
		progressBar.setPreferredSize(new Dimension(250, 20));

		label = new JLabel();
		label.setHorizontalAlignment(SwingConstants.LEFT);
		label.setVerticalAlignment(SwingConstants.BOTTOM);
		label.setBorder(new EmptyBorder(10, 15, 1, 15));
		label.setPreferredSize(new Dimension(250, 100));

		JPanel panel = new JPanel(new BorderLayout());
		panel.add(label, BorderLayout.CENTER);
		panel.add(progressBar, BorderLayout.SOUTH);

		dialog.add(progressBar);
		dialog.pack();
		dialog.setLocationRelativeTo(null);

		NamedRunnable[] hooks = Core.getCore().getShutdownHooks();
		progressBar.setMaximum(hooks.length);

		// Dispatch worker
		new ShutdownWorker(hooks).execute();

		dialog.setVisible(true);
	}

	private class ShutdownWorker extends SwingWorker<Object, String> {

		private final NamedRunnable[] hooks;

		ShutdownWorker(NamedRunnable[] hooks) {
			this.hooks = hooks;
		}

		/**
		 * @see javax.swing.SwingWorker#doInBackground()
		 */
		@Override
		protected Object doInBackground() throws Exception {

			for(NamedRunnable hook : hooks) {
				try {
					publish(hook.getName());
					hook.run();
				} catch(Exception e) {
					LoggerFactory.log(this, Level.SEVERE,
							"Failed to execute shutdown hook: "+hook.getName(), e); //$NON-NLS-1$
				}
			}

			return null;
		}

		@Override
		protected void process(List<String> chunks) {
			if(chunks==null || chunks.isEmpty()) {
				return;
			}

			String text = chunks.get(chunks.size()-1);
			int steps = chunks.size();

			label.setText(text);
			progressBar.setValue(progressBar.getValue()+steps);
		}

		@Override
		protected void done() {
			boolean success = true;
			try {
				get();
			} catch(Exception e) {
				success = false;
				LoggerFactory.log(this, Level.SEVERE,
						"Failed to execute all shutdown hooks", e); //$NON-NLS-1$
			}

			try {
				FrameManager.getInstance().shutdown();
			} catch(Exception e) {
				success = false;
				LoggerFactory.log(this, Level.SEVERE,
						"Failed to shutdown frame manager", e); //$NON-NLS-1$
			}

			System.exit(success ? 0 : 1);
		}

	}
}
