package ims;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Date;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import is2.data.SentenceData09;
import is2.io.CONLLReader09;
import is2.io.CONLLWriter09;
import is2.lemmatizer.Lemmatizer;
import is2.parser.Options;
import is2.parser.ParametersFloat;
import is2.parser.Parser;
import is2.parser.Pipe;
import is2.util.ParserEvaluator;

public class AnnaPipeline {

	final Lemmatizer lemmatizer;
	final is2.tag.Tagger tagger;
	final is2.mtag.Tagger mtagger;
	final Parser parser;
	final int newLineInterval;
	
	public AnnaPipeline(Lemmatizer lemmatizer,is2.tag.Tagger tagger,is2.mtag.Tagger mtagger,Parser parser,int newLineInterval) {
		this.lemmatizer=lemmatizer;
		this.tagger=tagger;
		this.mtagger=mtagger;
		this.parser=parser;
		this.newLineInterval=newLineInterval;
	}

	private static CONLLWriter09 getWriter(AnnaPipelineOptions options) throws IOException {
		OutputStream os=new FileOutputStream(options.out);
		if(options.out.toString().endsWith(".gz"))
			os=new GZIPOutputStream(os);
		OutputStreamWriter osw=new OutputStreamWriter(os,"UTF-8");
		CONLLWriter09 writer=new CONLLWriter09(osw);
		return writer;
	}

	private static CONLLReader09 getReader(AnnaPipelineOptions options) throws IOException {
		InputStream is=new FileInputStream(options.in);
		if(options.in.toString().endsWith(".gz"))
			is=new GZIPInputStream(is);
		CONLLReader09 reader=new CONLLReader09();
		reader.startReading(is);
		return reader;
	}

	public static AnnaPipeline getPipeline(File lemmaModel,File tagModel,File morphModel,File parseModel,int cores,boolean doUppercaseLemmas, int newLineInterval) throws IOException{
		DebugOut.println("Loading models");
		Lemmatizer lemmatizer=null;
		if(lemmaModel!=null)
			lemmatizer=new Lemmatizer(lemmaModel.getAbsolutePath(),doUppercaseLemmas);
		
		is2.tag.Tagger tagger=null;
		if(tagModel!=null)
			tagger=new is2.tag.Tagger(tagModel.getAbsolutePath());
		
		is2.mtag.Tagger mtag=null;
		if(morphModel!=null)
			mtag=new is2.mtag.Tagger(morphModel.getAbsolutePath());
		
		Parser parser=null;
		if(parseModel!=null){
			Options options=new is2.parser.Options(new String[]{"-model",parseModel.getAbsolutePath()});
			Parser.THREADS=cores;
			Parser p = new Parser();
			p.options=options;
			p. pipe = new Pipe(options);
			p. params = new ParametersFloat(0);
			p.readModel(options, p.pipe, p.params);
			parser=p;
		}
		return new AnnaPipeline(lemmatizer,tagger,mtag,parser,newLineInterval);
	}
	
	public void processParallel(CONLLReader09 reader, CONLLWriter09 writer) throws IOException {
		BlockingQueue<SentenceData09> queue=new ArrayBlockingQueue<SentenceData09>(10);
		PreprocessWorker w1=new PreprocessWorker(queue,reader);
		ParserWorker w2=new ParserWorker(queue,writer);
		long start=System.currentTimeMillis();
		w1.start();
		w2.start();
		DebugOut.out.print("Processing sentence ");
		int erase=0;
		int last=0;
		while(true){
			try {
				Thread.sleep(167);
			} catch (InterruptedException e) {}
			
			int out=w2.outCount;
			if(out==0)
				continue;
			long avgTime=(System.currentTimeMillis()-start)/out;
			erase=DebugOut.eraseAndPrint(erase, String.format("  %d     (%dms/s)", out,avgTime));
			if(out-last>=newLineInterval){
				DebugOut.out.println(" at "+new Date());
				erase=0;
				DebugOut.out.print("Processing sentence ");
				last=out-out%newLineInterval;
			}
			if(doneInput && doneOutput)
				break;
		}
	}
	
	
	public void process(CONLLReader09 reader, CONLLWriter09 writer) throws IOException {
		int sentCount=0;
		long aggregate=0;
		int erase=0;
		DebugOut.out.print("Processing sentence ");
		while(true){
			long start=System.currentTimeMillis();
			SentenceData09 sent=reader.getNext();
			if(sent==null)
				break;
			sentCount++;
			
			if(lemmatizer!=null)
				lemmatizer.apply(sent);
			if(tagger!=null)
				tagger.apply(sent);
			if(mtagger!=null){
				mtagger.apply(sent);
				if(parser!=null && sent.pfeats!=null){
					for(int i=1;i<sent.pfeats.length;++i){
						if(sent.pfeats[i]!=null && !sent.pfeats[i].equals("_"))
							sent.feats[i]=sent.pfeats[i].split("\\|");
					}
				}
			}
			if(parser!=null)
				sent=parser.applyQuick(sent);
			aggregate+=(System.currentTimeMillis()-start);

			if(parser==null){
				SentenceData09 out=new SentenceData09(sent);
				out.createSemantic(sent);
				writer.write(out, true);
			} else
				writer.write(sent);
			
			writer.flush();
			
			long avgTime=aggregate/sentCount;
			erase=DebugOut.eraseAndPrint(erase, String.format("  %d     (%dms/s)", sentCount,avgTime));
			if(sentCount%newLineInterval==0){ //Debug and newline
				DebugOut.out.println(" at "+new Date());
				erase=0;
				DebugOut.out.print("Processing sentence ");
			}
		}
	}
	
	public static void main(String[] args) throws Exception{
		AnnaPipelineOptions options=new AnnaPipelineOptions(args);
		AnnaPipeline pipeline=getPipeline(options.lemmaModel,options.tagModel,options.morphModel,options.parseModel,options.cores,options.doUppercaseLemmas,options.newLineInterval);
		CONLLReader09 reader=getReader(options);
		CONLLWriter09 writer=getWriter(options);
		DebugOut.out.println();
		DebugOut.out.println("Starting pipeline at "+new Date());
		if(pipeline.parser!=null)
			DebugOut.out.println("Using "+Parser.THREADS+" cores for the parser");
		DebugOut.out.println();
		pipeline.processParallel(reader,writer);
		writer.finishWriting();
		if(pipeline.parser!=null){
			is2.parser.Decoder.executerService.shutdown();
			is2.parser.Pipe.executerService.shutdown();
		}
		DebugOut.out.println();
		DebugOut.out.println("Done at "+new Date());
		if(options.eval!=null){
			evaluate(options);
		}
	}

	private static void evaluate(AnnaPipelineOptions options) throws Exception {
		System.out.println();
		System.out.println("Evaluation");
		System.out.println();
		if(options.lemmaModel!=null){
			System.out.println("Evaluation of LEMMATIZER");
			is2.lemmatizer.Evaluator.evaluate(options.eval.toString(), options.out.toString(), null);
			System.out.println("-------------------------");
		}
		if(options.tagModel!=null){
			System.out.println("Evaluation of TAGGER");
			is2.util.Evaluator.evaluateTagger(options.eval.toString(), options.out.toString(), "");
			System.out.println("-------------------------");
		}
		if(options.morphModel!=null){
			System.out.println("Evaluation of MTAGGER");
			is2.mtag.Evaluator.evaluate(options.eval.toString(), options.out.toString(), null);
			System.out.println("-------------------------");
		}
		if(options.parseModel!=null){
			System.out.println("Evaluation of PARSER");
			ParserEvaluator.evaluate(options.eval.toString(), options.out.toString());
			System.out.println("-------------------------");
		}
	}
	
	class ParserWorker extends Thread {
		final BlockingQueue<SentenceData09> queue;
		final CONLLWriter09 writer;
		int outCount=0;
		
		public ParserWorker(BlockingQueue<SentenceData09> queue,CONLLWriter09 writer){
			this.queue=queue;
			this.writer=writer;
		}
		
		public void run(){
			try {
				while(true){
					SentenceData09 sent;
					try {
						sent = queue.take();
					} catch (InterruptedException e) {
						continue;
					}
					if(sent.forms==null)
						break;
					
					if(parser!=null)
						sent=parser.applyQuick(sent);

					if(parser==null){
						SentenceData09 out=new SentenceData09(sent);
						out.createSemantic(sent);
						writer.write(out, true);
					} else
						writer.write(sent);
					
					writer.flush();
					outCount++;
				}
			} catch(IOException e){
				throw new RuntimeException(e);
			}
//			System.out.println("Done output");
			doneOutput=true;
		}
	}

	boolean doneInput=false;
	boolean doneOutput=false;
	
	class PreprocessWorker extends Thread {
		final BlockingQueue<SentenceData09> queue;
		final CONLLReader09 reader;

		
		public PreprocessWorker(BlockingQueue<SentenceData09> queue,CONLLReader09 reader){
			this.queue=queue;
			this.reader=reader;
		}
		
		public void run(){
			while(true){
				SentenceData09 sent=reader.getNext();
				if(sent==null)
					break;
				if(lemmatizer!=null)
					lemmatizer.apply(sent);
				if(tagger!=null)
					tagger.apply(sent);
				if(mtagger!=null){
					mtagger.apply(sent);
					if(parser!=null && sent.pfeats!=null){
						for(int i=1;i<sent.pfeats.length;++i){
							if(sent.pfeats[i]!=null && !sent.pfeats[i].equals("_"))
								sent.feats[i]=sent.pfeats[i].split("\\|");
						}
					}
				}
				try {
					queue.put(sent);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			doneInput=true;
			SentenceData09 sentinel=new SentenceData09();
			try {
				queue.put(sentinel);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
//			System.out.println("Done input");
		}
	}

}
