package ims.gw.tossduplicates;

import ims.DebugOut;
import ims.gw.tossduplicates.Options.Format;
import is2.data.SentenceData09;
import is2.io.CONLLReader09;
import is2.io.CONLLWriter09;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.HashSet;
import java.util.Set;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

public class TossDuplicates {

	public static void main(String[] args) throws IOException{
		Options options=new Options(args);
		if(options.input==null && options.inputFiles==null)
			Options.die("No input specified, aborting");

		if(options.input!=null){
			doIOOneFile(options);
		} else if(options.inputFiles!=null){
			doIOFileList(options);
		}
	}
	
	private int curFile=0;
	private int totalFiles=1;
	private int totalSeen=0;
	private int totalKept=0;
	private Set<String> seen;
	private String inputEnc;
	private String outputEnc;
	private Format inputFormat;
	private Format outputFormat;
	
	
	public TossDuplicates(Options options){
		seen=new HashSet<String>();
		this.inputEnc=options.inputEnc;
		this.outputEnc=options.outputEnc;
		this.inputFormat=options.inputFormat;
		this.outputFormat=options.outputFormat;
	}
	
	private void process(File inputFile,File outputFile) throws IOException{
		int seenThisTime=0;
		int keptThisTime=0;
		IInput input=getInput(inputFile);
		IOutput output=getOutput(outputFile);
		curFile++;
		
		String h=String.format("P: %s -> %s (%d / %d)  ", inputFile.toString(),outputFile.toString(),curFile,totalFiles);
		DebugOut.print(h);
		
		int erase=0;
		while(input.hasNext()){
			String s=input.getNextTokensAsString();
			totalSeen++;
			seenThisTime++;
			Object item=input.getNext();
			if(seen.contains(s))
				continue;
			seen.add(s);
			output.writeItem(item);
			totalKept++;
			keptThisTime++;
			if(seenThisTime%10000==0)
				erase=DebugOut.eraseAndPrint(erase, formatInfo(seenThisTime, keptThisTime));
		}
		DebugOut.eraseAndPrint(erase, formatInfo(seenThisTime, keptThisTime));
		DebugOut.println();
	}

	private String formatInfo(int seenThisTime, int keptThisTime) {
		double percTotal=100.0*totalKept/totalSeen;
		double percThis=100.0*keptThisTime/seenThisTime;
		return String.format("%6d / %6d (%.2f%%) | tot:  %10d / %10d (%.2f%%)", keptThisTime,seenThisTime,percThis,totalKept,totalSeen,percTotal);
	}
	
	private static void doIOFileList(Options options) throws IOException {
		TossDuplicates td=new TossDuplicates(options);
		td.totalFiles=options.inputFiles.length;
		for(File f:options.inputFiles){
			File out=new File(f.toString()+options.outputFilesSuffix);
			td.process(f, out);
		}
	}




	private static void doIOOneFile(Options options) throws IOException {
		if(options.output==null)
			Options.die("No output file specified, aborting.");
		
		TossDuplicates td=new TossDuplicates(options);
		td.process(options.input,options.output);
	}




	private IOutput getOutput(File output) throws IOException {
		switch(outputFormat){
		case OneSentencePerLine: return new OneSentencePerLineOutput(output,outputEnc);
		case CoNLL09:            return new CoNLL09Output(output,outputEnc);
		default: throw new Error("not implemented");
		}
	}

	private IInput getInput(File input) throws IOException {
		switch(inputFormat){
		case OneSentencePerLine: return new OneSentencePerLineInput(input,inputEnc);
		case CoNLL09:            return new CoNLL09Input(input,inputEnc);
		default: throw new Error("not implemented");
		}
	}




	private static interface IInput {
		public String getNextTokensAsString();
		public Object getNext() throws IOException;
		public boolean hasNext();
	}
	
	private static class CoNLL09Input implements IInput {

		private CONLLReader09 reader;
		private SentenceData09 next;
		
		public CoNLL09Input(File input, String inputEnc) throws IOException {
			InputStream is=new FileInputStream(input);
			if(input.toString().endsWith(".gz"))
				is=new GZIPInputStream(is);
			reader=new CONLLReader09();
			reader.startReading(is,inputEnc);
			next=reader.getNextCoNLL09();
		}

		@Override
		public String getNextTokensAsString() {
			return next.oneLine();
		}

		@Override
		public Object getNext() throws IOException {
			SentenceData09 n=next;
			next=reader.getNextCoNLL09();
			return n;
		}

		@Override
		public boolean hasNext() {
			return next!=null;
		}
		
	}
	
	private static class OneSentencePerLineInput implements IInput {
		
		private String next;
		private BufferedReader reader;
		
		public OneSentencePerLineInput(File input,String encoding) throws IOException{
			InputStream is=new FileInputStream(input);
			if(input.toString().endsWith(".gz"))
				is=new GZIPInputStream(is);
			this.reader=new BufferedReader(new InputStreamReader(is,encoding));
			next=reader.readLine();
		}
		
		@Override
		public String getNextTokensAsString() {
			return next;
		}

		@Override
		public boolean hasNext() {
			return next!=null;
		}

		@Override
		public Object getNext() throws IOException {
			String n=next;
			String line=reader.readLine();
			next=line;
			if(next==null)
				reader.close();
			return n;
		}
	}
	
	private static interface IOutput {
		public void writeItem(Object item) throws IOException;
		public void close() throws IOException;
	}
	
	private static class CoNLL09Output implements IOutput {

		private CONLLWriter09 out;
		
		public CoNLL09Output(File output, String outputEnc) throws IOException {
			OutputStream os=new FileOutputStream(output);
			if(output.toString().endsWith(".gz"))
				os=new GZIPOutputStream(os);
			OutputStreamWriter osw=new OutputStreamWriter(os,"UTF-8");
			out=new CONLLWriter09(osw);
		}

		@Override
		public void writeItem(Object item) throws IOException {
			if(item instanceof String){
				String[] tokens=((String) item).split(" ");
//				String[] t=new String[tokens.length+1];
//				t[0]=IOGenerals.ROOT;
//				System.arraycopy(tokens, 0, t, 1, tokens.length);
				SentenceData09 s09=new SentenceData09();
				s09.init(tokens);
				out.write(s09, false);
			} else if(item instanceof SentenceData09){
				out.write((SentenceData09) item,true);
			} else {
				throw new RuntimeException("!");
			}
		}

		@Override
		public void close() throws IOException {
			out.finishWriting();
		}
		
	}

	private static class OneSentencePerLineOutput implements IOutput {

		private BufferedWriter out;
		
		public OneSentencePerLineOutput(File output, String outputEnc) throws IOException {
			OutputStream os=new FileOutputStream(output);
			if(output.toString().endsWith(".gz"))
				os=new GZIPOutputStream(os);
			out=new BufferedWriter(new OutputStreamWriter(os,outputEnc));
		}

		@Override
		public void writeItem(Object item) throws IOException {
			if(item instanceof String){
				out.write(item.toString());
				out.write('\n');
			} else if(item instanceof SentenceData09){
				SentenceData09 s09=(SentenceData09) item;
				StringBuilder sb=new StringBuilder(s09.forms[1]);
				for(int i=2;i<s09.forms.length;++i)
					sb.append(' ').append(s09.forms[i]);
				out.write(sb.toString());
				out.write('\n');
			} else {
				throw new RuntimeException("!");
			}
		}

		@Override
		public void close() throws IOException {
			out.flush();
			out.close();
		}
		
	}
}
