package is2.parser;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;

import is2.data.DataFES;
import is2.data.Parse;
import is2.data.ParseNBest;
import is2.parser.ParallelRearrangeNBestD8.RearrangeJob;
import is2.parser.ParallelRearrangeNBestD8.SyncJobQueue;
import is2.parserR2.Parser;
import is2.util.DB;


/**
 * @author Bernd Bohnet, 01.09.2009
 * 
 * This methods do the actual work and they build the dependency trees. 
 */
final public class Decoder   {

	public static final boolean TRAINING = true;
	public static long timeDecotder;
	public static long timeRearrange;

	/**
	 * Threshold for rearrange edges non-projective
	 */
	public static float NON_PROJECTIVITY_THRESHOLD = 0.3F;

	
	public static ExecutorService executerService =java.util.concurrent.Executors.newFixedThreadPool(Parser.THREADS);

	// do not initialize
	private Decoder() {};
	
	
	/**
	 * Build a dependency tree based on the data
	 * @param pos part-of-speech tags
	 * @param x the data
	 * @param projective projective or non-projective
	 * @param edges the edges
	 * @return a parse tree
	 * @throws InterruptedException
	 */
	public static Parse  decode(short[] pos,  DataFES x, boolean projective, boolean training) throws InterruptedException {

		long ts = System.nanoTime();
		
		if (executerService.isShutdown()) executerService = java.util.concurrent.Executors.newCachedThreadPool();
		final int n = pos.length;

		final Open O[][][][] = new Open[n][n][2][];
		final Closed C[][][][] = new Closed[n][n][2][];

		ArrayList<ParallelDecoder> pe = new ArrayList<ParallelDecoder>(); 

		for(int i=0;i<Parser.THREADS ;i++)  pe.add(new ParallelDecoder(pos, x,  O, C, n));
		
		for (short k = 1; k < n; k++) {

			// provide the threads the data
			for (short s = 0; s < n; s++) {
				short t = (short) (s + k);
				if (t >= n) break;
				
				ParallelDecoder.add(s,t);
			}
						
			executerService.invokeAll(pe);
		}
		
		float bestSpanScore = (-1.0F / 0.0F);
		Closed bestSpan = null;
		for (int m = 1; m < n; m++)
			if (C[0][n - 1][1][m].p > bestSpanScore) {
				bestSpanScore = C[0][n - 1][1][m].p;
				bestSpan = C[0][n - 1][1][m];
			}

		// build the dependency tree from the chart 
		Parse out= new Parse(pos.length);

		bestSpan.create(out);

		out.heads[0]=-1;
		out.labels[0]=0;

		timeDecotder += (System.nanoTime()-ts);
		 
		ts = System.nanoTime();
		
		if (!projective) rearrange(pos, out.heads, out.labels,x,training);
		
		timeRearrange += (System.nanoTime()-ts);		

		return out;
	}


	public static List<? extends Parse> decodeNBest(short[] pos, DataFES x,boolean projective, boolean b,int nbest) throws InterruptedException, ExecutionException {
		long ts = System.nanoTime();
		
		if (executerService.isShutdown()) executerService = java.util.concurrent.Executors.newCachedThreadPool();
		final int n = pos.length;

		final Open O[][][][] = new Open[n][n][2][];
		final Closed C[][][][] = new Closed[n][n][2][];

		ArrayList<ParallelDecoder> pe = new ArrayList<ParallelDecoder>(); 

		for(int i=0;i<Parser.THREADS ;i++)  pe.add(new ParallelDecoder(pos, x,  O, C, n));
		
		for (short k = 1; k < n; k++) {

			// provide the threads the data
			for (short s = 0; s < n; s++) {
				short t = (short) (s + k);
				if (t >= n) break;
				
				ParallelDecoder.add(s,t);
			}
						
			executerService.invokeAll(pe);
		}
		
		float bestSpanScore = (-1.0F / 0.0F);
		Closed bestSpan = null;
		for (int m = 1; m < n; m++)
			if (C[0][n - 1][1][m].p > bestSpanScore) {
				bestSpanScore = C[0][n - 1][1][m].p;
				bestSpan = C[0][n - 1][1][m];
			}

		// build the dependency tree from the chart 
		Parse out= new Parse(pos.length);

		bestSpan.create(out);

		out.heads[0]=-1;
		out.labels[0]=0;

		Parse bestProj=out;
		timeDecotder += (System.nanoTime()-ts);
		//	DB.println(""+out);
		
		ts = System.nanoTime();
		List<? extends Parse> parses;
		if (!projective) {
			parses = rearrangeNBestD8(pos, out.heads, out.labels,x,nbest);
		} else {
			throw new Error("nbest only for non-projective");
		}
		
		timeRearrange += (System.nanoTime()-ts);		

		return parses;
	}
	
	private static List<? extends Parse> rearrangeNBestD8(short[] pos,short[] heads_,short[] labs_, DataFES x,int n) throws InterruptedException, ExecutionException {
		return D8NBestByRearranging.rearrangeNBest(pos,heads_,labs_,x,n);
	}
//	
//	private static List<ParseNBest> rearrangeNBest(short[] pos, short[] heads, short[] labs, DataFES x,int n) throws InterruptedException {
//		ArrayList<ParallelRearrangeNBest> pe = new ArrayList<ParallelRearrangeNBest>();
//
//		int round =0;
//		ArrayList<ParseNBest> parses = new ArrayList<ParseNBest>();
//		ParseNBest px =new ParseNBest();
//		px.signature(heads,labs);
//
//
//		px.f1=Extractor.encode3(pos, heads, labs, x);
//		parses.add(px);
//
//		float lastNBest = Float.NEGATIVE_INFINITY;
//
//		HashSet<Parse> done = new HashSet<Parse>();
//		gnu.trove.THashSet<CharSequence> contained = new gnu.trove.THashSet<CharSequence>();
//
//		while(true) {
//			
//			pe.clear();
//
//			// used the first three parses
//			int i=0;
//			while(true) {
//				
//				if (parses.size()<=i||pe.size()>12) break;
//				
//				ParseNBest parse = parses.get(i);
//				
//				i++;
//				
//				// parse already extended
//				if (done.contains(parse))  continue;
//
////				DB.println("err "+parse.heads);
//
//				parse.signature2parse(parse.signature());
//				
//				done.add(parse);
//				pe.add(new ParallelRearrangeNBest( pos,x,parse,lastNBest, (float)parse.f1,NON_PROJECTIVITY_THRESHOLD) );
//		
//		
//			executerService.invokeAll(pe);
//				
//			// avoid to add parses several times
//			for(ParallelRearrangeNBest rp : pe) {
//				for(int k=rp.parses.size()-1;k>=0;k--) {
//					if (lastNBest>rp.parses.get(k).f1) continue;
//					CharSequence sig = rp.parses.get(k).signature();
//					if (!contained.contains(sig)) {
//						parses.add(rp.parses.get(k));
//						contained.add(sig);
//					}
//				}
//			}
//			}	
//			Collections.sort(parses);
//			
//			if (round >=2) break;
//			round ++;
//					
//			// do not use to much memory
//			if (parses.size()>n) {
//				if (parses.get(n).f1>lastNBest) lastNBest = (float)parses.get(n).f1;
//				parses.subList(n, parses.size()-1).clear();
//			}
//		}
//		return parses;
//	}


	/**
	 * This is the parallel non-projective edge re-arranger
	 *  
	 * @param pos part-of-speech tags
	 * @param heads parent child relation 
	 * @param labs edge labels 
	 * @param x the data
	 * @param edges the existing edges defined by part-of-speech tags 
	 * @throws InterruptedException
	 */
	public static void rearrange(short[] pos, short[] heads, short[] labs,  DataFES x, boolean training) throws InterruptedException {

		int threads =(pos.length>Parser.THREADS)? Parser.THREADS: pos.length;

		
		
		// wh  what to change, nPar - new parent, nType - new type
		short wh = -1, nPar = -1,nType = -1;
		ArrayList<ParallelRearrange> pe = new ArrayList<ParallelRearrange>(); 
		
		while(true) {
			boolean[][] isChild = new boolean[heads.length][heads.length];
			for(int i = 1, l1=1; i < heads.length; i++,l1=i)  
				while((l1= heads[l1]) != -1) isChild[l1][i] = true;
		
			float max = Float.NEGATIVE_INFINITY;
			float p = Extractor.encode3(pos, heads, labs, x);

			pe.clear();
			for(int i=0;i<threads;i++)  pe.add(new ParallelRearrange( isChild, pos,x,heads,labs));
			
			for(int ch = 1; ch < heads.length; ch++) {

				for(short pa = 0; pa < heads.length; pa++) {
					if(ch == pa || pa == heads[ch] || isChild[ch][pa]) continue;

					ParallelRearrange.add(p,(short) ch, pa);
				} 
			}  
			executerService.invokeAll(pe);
			
			for(ParallelRearrange.PA rp :ParallelRearrange.order) 
				if(max < rp.max  ) {					
						max = rp.max;  	wh = rp.wh; 
						nPar = rp.nPar;  nType = rp.nType ;
				}
			ParallelRearrange.order.clear();
			
			if(max <= NON_PROJECTIVITY_THRESHOLD)	break; // bb: changed from 0.0

			heads[wh] = nPar;
			labs[wh] = nType;

		}
	}

	public static String getInfo() {

		return "Decoder non-projectivity threshold: "+NON_PROJECTIVITY_THRESHOLD;
	}




}
