/*
 *  ICARUS -  Interactive platform for Corpus Analysis and Research tools, University of Stuttgart
 *  Copyright (C) 2012-2013 Markus Gärtner and Gregor Thiele
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses.
 *
 * $Revision: 389 $
 * $Date: 2015-04-23 12:19:15 +0200 (Do, 23 Apr 2015) $
 * $URL: https://subversion.assembla.com/svn/icarusplatform/trunk/Icarus/plugins/de.ims.icarus.plugins.core/source/de/ims/icarus/plugins/core/output/DefaultOutputView.java $
 *
 * $LastChangedDate: 2015-04-23 12:19:15 +0200 (Do, 23 Apr 2015) $
 * $LastChangedRevision: 389 $
 * $LastChangedBy: mcgaerty $
 */
package de.ims.icarus.plugins.core.output;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.logging.Level;

import javax.swing.Icon;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;

import de.ims.icarus.Core;
import de.ims.icarus.config.ConfigRegistry;
import de.ims.icarus.logging.LoggerFactory;
import de.ims.icarus.plugins.core.ManagementConstants;
import de.ims.icarus.plugins.core.View;
import de.ims.icarus.resources.ResourceManager;
import de.ims.icarus.ui.UIUtil;
import de.ims.icarus.ui.events.EventListener;
import de.ims.icarus.ui.events.EventObject;
import de.ims.icarus.ui.helper.UIHelperRegistry;
import de.ims.icarus.ui.text.Console;
import de.ims.icarus.ui.view.AWTPresenter;
import de.ims.icarus.ui.view.PresenterUtils;
import de.ims.icarus.ui.view.UnsupportedPresentationDataException;
import de.ims.icarus.util.Options;
import de.ims.icarus.util.Wrapper;
import de.ims.icarus.util.mpi.Commands;
import de.ims.icarus.util.mpi.Message;
import de.ims.icarus.util.mpi.ResultMessage;


/**
 * @author Markus Gärtner
 * @version $Id: DefaultOutputView.java 389 2015-04-23 10:19:15Z mcgaerty $
 *
 */
public class DefaultOutputView extends View implements ManagementConstants {

	public static final String VIEW_ID = ManagementConstants.DEFAULT_OUTPUT_VIEW_ID;

	private JTabbedPane presenterPane;

	private Handler handler;

	public DefaultOutputView() {
		// no-op
	}

	/**
	 * @see de.ims.icarus.plugins.core.View#init(javax.swing.JComponent)
	 */
	@Override
	public void init(JComponent container) {
		handler = new Handler();

		addBroadcastListener(ManagementConstants.EXPLORER_SELECTION_CHANGED, handler);
		addBroadcastListener(ManagementConstants.LOG_SELECTION_CHANGED, handler);

		if(ConfigRegistry.getGlobalRegistry().getBoolean(
				"general.appearance.showSystemStreams") //$NON-NLS-1$
				&& !Core.ignoreRedirect()) {
			Console console = Console.getInstance();

			// Display sys.out redirect
			Options options = new Options();
			options.put(Options.TITLE, "System.out"); //$NON-NLS-1$
			displayData(console.getOutputDocument(), console, options);

			if(!console.isMergeStreams()) {
				// Display sys.err redirect
				options = new Options();
				options.put(Options.TITLE, "System.err"); //$NON-NLS-1$
				displayData(console.getErrorDocument(), console, options);
			}
		}
	}

	/**
	 * @see de.ims.icarus.plugins.core.View#close()
	 */
	@Override
	public void close() {
		reset();
	}

	/**
	 * @see de.ims.icarus.plugins.core.View#reset()
	 */
	@Override
	public void reset() {
		if(presenterPane==null) {
			return;
		}

		for(int i=presenterPane.getTabCount()-1; i>-1; i--) {
			TabComponent tabComponent = (TabComponent) presenterPane.getTabComponentAt(i);
			if(tabComponent.presenter!=null) {
				tabComponent.presenter.close();
			}
		}
		presenterPane.removeAll();
	}

	private TabComponent findTabForOwner(Object owner) {
		if(owner==null) {
			return null;
		}

		TabComponent comp = null;

		for(int i=0; i<presenterPane.getTabCount(); i++) {
			comp = (TabComponent)presenterPane.getTabComponentAt(i);
			if(comp.getOwner()==owner) {
				return comp;
			}
		}

		return null;
	}

	private boolean canDisplay(Object data) {
		if(data==null) {
			return false;
		}
		return UIHelperRegistry.globalRegistry().hasHelper(
				AWTPresenter.class, data);
	}

	public void displayData(Object data, Object owner, Options options) {
		if(data==null)
			throw new NullPointerException("Invalid data"); //$NON-NLS-1$
		if(owner==null)
			throw new NullPointerException("Invalid owner"); //$NON-NLS-1$

		selectViewTab();

		if(options==null) {
			options = Options.emptyOptions;
		}

		// Lazily create presenter pane
		if(presenterPane==null) {
			presenterPane = new JTabbedPane();
			UIUtil.defaultHideTabbedPaneDecoration(presenterPane);
			presenterPane.setBorder(null);
			JComponent container = getContainer();
			container.setLayout(new BorderLayout());
			container.add(presenterPane, BorderLayout.CENTER);
		}

		TabComponent tabComponent = null;

		// Recycle tabs if allowed
		if(options.get(ManagementConstants.REUSE_TAB_OPTION, false)) {
			tabComponent = findTabForOwner(owner);
		}

		int tabIndex = -1;
		if(tabComponent==null) {
			tabComponent = new TabComponent(owner);
			tabIndex = presenterPane.getTabCount();
			presenterPane.add(new JLabel(), tabIndex);
			presenterPane.setTabComponentAt(tabIndex, tabComponent);
		} else {
			tabIndex = presenterPane.indexOfTabComponent(tabComponent);
		}

		tabComponent.setPresentedObject(data, options);

		presenterPane.setSelectedIndex(tabIndex);
	}

	/**
	 * Accepted commands:
	 * <ul>
	 * <li>{@link Commands#DISPLAY}</li>
	 * <li>{@link Commands#PRESENT}</li>
	 * <li>{@link Commands#CLEAR}</li>
	 * </ul>
	 *
	 * @see de.ims.icarus.plugins.core.View#handleRequest(de.ims.icarus.util.mpi.Message)
	 */
	@Override
	protected ResultMessage handleRequest(Message message) throws Exception {

		Object data = message.getData();

		// Unwrap wrapped data
		if(data instanceof Wrapper) {
			data = ((Wrapper<?>)data).get();
		}

		Object owner = message.getOption(Options.OWNER);

		if(Commands.DISPLAY.equals(message.getCommand())
				|| Commands.PRESENT.equals(message.getCommand())) {

			if(owner==null) {
				return message.errorResult(this, new IllegalArgumentException("Missing owner")); //$NON-NLS-1$
			}

			// Check if there is a presenter available for the supplied object
			if(!canDisplay(data)) {
				return message.unsupportedDataResult(this);
			}

			displayData(data, owner, message.getOptions());

			return message.successResult(this, null);
		} else if(Commands.CLEAR.equals(message.getCommand())) {
			reset();
			return message.successResult(this, null);
		} else {
			return message.unknownRequestResult(this);
		}
	}

	/**
	 *
	 * @author Markus Gärtner
	 * @version $Id: DefaultOutputView.java 389 2015-04-23 10:19:15Z mcgaerty $
	 *
	 */
	private class TabComponent extends JLabel implements MouseListener {

		private static final long serialVersionUID = -9157665763570448381L;

		private AWTPresenter presenter;
		@SuppressWarnings("unused")
		private Object presentedObject;
		private Options options;
		private final Object owner;

		TabComponent(Object owner) {
			this.owner = owner;
			addMouseListener(this);
		}

		@SuppressWarnings("unused")
		private void close() {
			if(presenter!=null) {
				presenter.clear();
			}

			presenter = null;
			presentedObject = null;
			options = null;
		}

		@Override
		public String getText() {
			int tabIndex = presenterPane.indexOfTabComponent(this);
			return tabIndex==-1 ? null : presenterPane.getTitleAt(tabIndex);
		}

		@Override
		public Icon getIcon() {
			int tabIndex = presenterPane.indexOfTabComponent(this);
			return tabIndex==-1 ? null : presenterPane.getIconAt(tabIndex);
		}

		@Override
		public String getToolTipText() {
			int tabIndex = presenterPane.indexOfTabComponent(this);
			return tabIndex==-1 ? null : presenterPane.getToolTipTextAt(tabIndex);
		}

		/**
		 * @return the owner
		 */
		public Object getOwner() {
			return owner;
		}

		private void showInfoLabel(String infoText) {
			int tabIndex = presenterPane.indexOfTabComponent(this);
			if(tabIndex==-1) {
				return;
			}
			JLabel infoLabel = new JLabel(infoText);
			infoLabel.setBorder(new EmptyBorder(50, 100, 100, 100));
			infoLabel.setVerticalAlignment(SwingConstants.TOP);
			infoLabel.setHorizontalAlignment(SwingConstants.CENTER);

			presenterPane.setComponentAt(tabIndex, infoLabel);
		}

		/**
		 * @param presentedObject the presentedObject to set
		 */
		public void setPresentedObject(Object presentedObject, Options options) {

			if(options==null) {
				options = Options.emptyOptions;
			}

			// Show info for empty content data
			if(presentedObject==null) {
				String infoText = ResourceManager.getInstance().get(
						"plugins.core.outputView.emptyContent"); //$NON-NLS-1$
				showInfoLabel(infoText);
				return;
			}

			AWTPresenter presenter = this.presenter;

			// Try to find a suitable presenter if no presenter
			// is present or the current presenter is not able to
			// present the given data
			if(presenter==null || !PresenterUtils.presenterSupports(presenter, presentedObject)) {
				presenter = UIHelperRegistry.globalRegistry().findHelper(
						AWTPresenter.class, presentedObject);
			}

			// Only with a suitable presenter show the content
			if(presenter!=null && PresenterUtils.presenterSupports(presenter, presentedObject)) {
				this.presentedObject = presentedObject;
				this.options = options;
				this.presenter = presenter;

				try {
					presenter.present(presentedObject, options);

					// Display presenter component
					int tabIndex = presenterPane.indexOfTabComponent(this);
					Component comp = presenter.getPresentingComponent();
					if(comp instanceof JComponent) {
						((JComponent)comp).setBorder(UIUtil.topLineBorder);
					}
					presenterPane.setComponentAt(tabIndex, comp);
					String title = options.get(Options.TITLE, String.valueOf(presentedObject));
					presenterPane.setTitleAt(tabIndex, title);
					presenterPane.setToolTipTextAt(tabIndex, title);
				} catch (UnsupportedPresentationDataException e) {
					LoggerFactory.log(this, Level.SEVERE, "Failed to present data: "+presentedObject, e); //$NON-NLS-1$

					// Show info if data not supported
					String infoText = ResourceManager.getInstance().get(
							"plugins.core.outputView.presentationFailed"); //$NON-NLS-1$
					showInfoLabel(infoText);
					return;
				}
			} else {
				// Inform user about unsupported data
				String infoText = ResourceManager.getInstance().get(
						"plugins.core.outputView.unsupportedContent"); //$NON-NLS-1$
				showInfoLabel(infoText);
			}
		}

		/**
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		@Override
		public void mouseClicked(MouseEvent e) {
			if(SwingUtilities.isLeftMouseButton(e) && e.getClickCount()==2) {
				DefaultOutputView.this.toggleContainer();
			}
		}

		/**
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		@Override
		public void mousePressed(MouseEvent e) {
			// no-op
		}

		/**
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		@Override
		public void mouseReleased(MouseEvent e) {
			// no-op
		}

		/**
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		@Override
		public void mouseEntered(MouseEvent e) {
			// no-op
		}

		/**
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		@Override
		public void mouseExited(MouseEvent e) {
			// no-op
		}
	}

	/**
	 *
	 * @author Markus Gärtner
	 * @version $Id: DefaultOutputView.java 389 2015-04-23 10:19:15Z mcgaerty $
	 *
	 */
	public final class CallbackHandler {

		private CallbackHandler() {
			// no-op
		}
	}

	/**
	 *
	 * @author Markus Gärtner
	 * @version $Id: DefaultOutputView.java 389 2015-04-23 10:19:15Z mcgaerty $
	 *
	 */
	private class Handler implements EventListener {

		/**
		 * @see de.ims.icarus.ui.events.EventListener#invoke(java.lang.Object, de.ims.icarus.ui.events.EventObject)
		 */
		@Override
		public void invoke(Object sender, EventObject event) {
			Object item = event.getProperty("item"); //$NON-NLS-1$

			if(!canDisplay(item)) {
				return;
			}

			Options options = (Options)event.getProperty("options"); //$NON-NLS-1$
			if(options==null) {
				options = Options.emptyOptions;
			}

			Object owner = options.get(Options.OWNER);
			if(owner==null) {
				LoggerFactory.log(this, Level.INFO, "No owner set for item to display: "+item); //$NON-NLS-1$
				return;
			}

			displayData(item, owner, options);
		}
	}
}
