/*
 *  ICARUS -  Interactive platform for Corpus Analysis and Research tools, University of Stuttgart
 *  Copyright (C) 2012-2013 Markus Gärtner and Gregor Thiele
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses.
 *
 * $Revision: 329 $
 * $Date: 2014-12-13 13:57:36 +0100 (Sa, 13 Dez 2014) $
 * $URL: https://subversion.assembla.com/svn/icarusplatform/trunk/Icarus/plugins/de.ims.icarus.plugins.core/source/de/ims/icarus/plugins/core/output/TableView.java $
 *
 * $LastChangedDate: 2014-12-13 13:57:36 +0100 (Sa, 13 Dez 2014) $
 * $LastChangedRevision: 329 $
 * $LastChangedBy: mcgaerty $
 */
package de.ims.icarus.plugins.core.output;

import java.awt.BorderLayout;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTextArea;

import de.ims.icarus.plugins.core.View;
import de.ims.icarus.resources.ResourceManager;
import de.ims.icarus.ui.UIDummies;
import de.ims.icarus.ui.UIUtil;
import de.ims.icarus.ui.helper.Outline;
import de.ims.icarus.ui.helper.UIHelperRegistry;
import de.ims.icarus.ui.view.AWTPresenter.TableBasedPresenter;
import de.ims.icarus.ui.view.PresenterUtils;
import de.ims.icarus.ui.view.UnsupportedPresentationDataException;
import de.ims.icarus.util.Options;
import de.ims.icarus.util.data.ContentType;
import de.ims.icarus.util.mpi.Commands;
import de.ims.icarus.util.mpi.Message;
import de.ims.icarus.util.mpi.ResultMessage;


/**
 * @author Markus Gärtner
 * @version $Id: TableView.java 329 2014-12-13 12:57:36Z mcgaerty $
 *
 */
public class TableView extends View implements Outline {

	private TableBasedPresenter presenter;

	protected JTextArea infoLabel;
	protected JPanel contentPanel;

	public TableView() {
		// no-op
	}

	/**
	 * @see de.ims.icarus.plugins.core.View#init(javax.swing.JComponent)
	 */
	@Override
	public void init(JComponent container) {

		container.setLayout(new BorderLayout());

		infoLabel = UIUtil.defaultCreateInfoLabel(container);
		container.add(infoLabel, BorderLayout.NORTH);

		contentPanel = new JPanel(new BorderLayout());
		container.add(contentPanel, BorderLayout.CENTER);

		showInfo(null);
	}

	public TableBasedPresenter getPresenter() {
		return presenter;
	}

	@Override
	public void close() {
		super.close();

		if(presenter!=null) {
			presenter.close();
			presenter = null;
		}
	}

	@Override
	public void reset() {
		if(contentPanel==null) {
			return;
		}

		displayData(null, null);
	}

	protected void showInfo(String text) {
		if(text==null) {
			text = ResourceManager.getInstance().get(
					"plugins.core.tableView.notAvailable"); //$NON-NLS-1$
		}
		infoLabel.setText(text);

		infoLabel.setVisible(true);
		contentPanel.setVisible(false);
		contentPanel.removeAll();

		// Close any active presenter and discard its reference
		if(presenter!=null) {
			presenter.close();
			presenter = null;
		}
	}

	protected void setPresenter(TableBasedPresenter presenter) {
		if(this.presenter==presenter) {
			return;
		}

		if(this.presenter!=null) {
			this.presenter.close();
		}

		this.presenter = presenter;

		if(this.presenter==null) {
			showInfo(null);
		}
	}

	protected void displayData(Object data, Options options) {

		// Show default info if nothing available to be displayed
		if(data==null) {
			showInfo(null);
			return;
		}

		if(options==null) {
			options = Options.emptyOptions;
		}

		// Fetch content type for data
		TableBasedPresenter presenter = this.presenter;

		// Fetch new presenter
		if(presenter==null || !PresenterUtils.presenterSupports(presenter, data)) {
			ContentType contentType = (ContentType) options.get(Options.CONTENT_TYPE);
			if(contentType!=null) {
				presenter = UIHelperRegistry.globalRegistry().findHelper(
						TableBasedPresenter.class, contentType, true, false);
			} else {
				presenter = UIHelperRegistry.globalRegistry().findHelper(
						TableBasedPresenter.class, data);
			}
		}

		// Abort if presenter not available for content type
		if(presenter==null) {
			String text = ResourceManager.getInstance().get(
					"plugins.core.tableView.unsupportedType", data.getClass()); //$NON-NLS-1$
			showInfo(text);
			return;
		}

		selectViewTab();

		boolean refreshPresentingComponent = presenter!=this.presenter;

		setPresenter(presenter);

		try {
			presenter.present(data, options);
		} catch (UnsupportedPresentationDataException e) {
			String text = ResourceManager.getInstance().get(
					"plugins.core.tableView.presentationFailed", data.getClass()); //$NON-NLS-1$
			showInfo(text);
			UIDummies.createDefaultErrorOutput(contentPanel, e);
			refreshPresentingComponent = false;
			return;
		}

		if(refreshPresentingComponent) {
			contentPanel.removeAll();
			contentPanel.add(presenter.getPresentingComponent(), BorderLayout.CENTER);
		}

		infoLabel.setVisible(false);
		contentPanel.setVisible(true);
	}

	/**
	 * Accepted commands:
	 * <ul>
	 * <li>{@link Commands#DISPLAY}</li>
	 * <li>{@link Commands#PRESENT}</li>
	 * <li>{@link Commands#CLEAR}</li>
	 * </ul>
	 *
	 * @see de.ims.icarus.plugins.core.View#handleRequest(de.ims.icarus.util.mpi.Message)
	 */
	@Override
	protected ResultMessage handleRequest(Message message) throws Exception {

		if(Commands.CLEAR.equals(message.getCommand())) {
			reset();
			return message.successResult(this, null);
		} else if(Commands.PRESENT.equals(message.getCommand())
				|| Commands.DISPLAY.equals(message.getCommand())) {
			displayData(message.getData(), message.getOptions());
			return message.successResult(this, null);
		} else {
			return message.unknownRequestResult(this);
		}
	}
}
