/*
 *  ICARUS -  Interactive platform for Corpus Analysis and Research tools, University of Stuttgart
 *  Copyright (C) 2012-2013 Markus Gärtner and Gregor Thiele
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses.
 *
 * $Revision: 189 $
 * $Date: 2013-12-04 16:37:42 +0100 (Mi, 04 Dez 2013) $
 * $URL: https://subversion.assembla.com/svn/icarusplatform/trunk/Icarus/plugins/de.ims.icarus.plugins.language_tools/source/de/ims/icarus/language/treebank/TreebankListDelegate.java $
 *
 * $LastChangedDate: 2013-12-04 16:37:42 +0100 (Mi, 04 Dez 2013) $ 
 * $LastChangedRevision: 189 $ 
 * $LastChangedBy: mcgaerty $
 */
package de.ims.icarus.language.treebank;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.List;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.ims.icarus.io.Loadable;
import de.ims.icarus.language.AvailabilityObserver;
import de.ims.icarus.language.DataType;
import de.ims.icarus.language.SentenceData;
import de.ims.icarus.language.SentenceDataList;
import de.ims.icarus.util.NamedObject;
import de.ims.icarus.util.data.ContentType;


/**
 * @author Markus Gärtner
 * @version $Id: TreebankListDelegate.java 189 2013-12-04 15:37:42Z mcgaerty $
 *
 */
public class TreebankListDelegate implements SentenceDataList, NamedObject, Loadable {

	private Reference<Treebank> ref;
	
	private List<ChangeListener> changeListeners;
	
	private OwnedChangeListener ownedChangeListener;
	
	public TreebankListDelegate(Treebank treebank) {
		if(treebank==null)
			throw new NullPointerException("Invalid treebank"); //$NON-NLS-1$
		
		setTreebank(treebank);
	}
	
	void setTreebank(Treebank treebank) {
		if(ref!=null && ref.get()==treebank) {
			return;
		}
		ref = new WeakReference<>(treebank);
		
		if(ownedChangeListener!=null) {
			ownedChangeListener.destroy();
		}
		
		ownedChangeListener = new OwnedChangeListener(this);
		treebank.addChangeListener(ownedChangeListener);
		
		fireChangeEvent();
	}
	
	public Treebank getTreebank() {
		Treebank treebank = ref.get();
		if(treebank==null) {
			return null;
		}
		
		return treebank;
	}
	
	public boolean hasTreebank() {
		return ref.get()!=null;
	}

	/**
	 * @see de.ims.icarus.util.data.DataList#size()
	 */
	@Override
	public int size() {
		Treebank treebank = getTreebank();
		return treebank==null ? 0 : treebank.size();
	}

	/**
	 * @see de.ims.icarus.util.data.DataList#get(int)
	 */
	@Override
	public SentenceData get(int index) {
		Treebank treebank = getTreebank();
		return treebank==null ? null : treebank.get(index);
	}

	/**
	 * @see de.ims.icarus.util.data.DataList#getContentType()
	 */
	@Override
	public ContentType getContentType() {
		Treebank treebank = getTreebank();
		return treebank==null ? null : treebank.getContentType();
	}

	/**
	 * @see de.ims.icarus.util.data.DataList#addChangeListener(javax.swing.event.ChangeListener)
	 */
	@Override
	public void addChangeListener(ChangeListener listener) {
		if(changeListeners==null) {
			changeListeners = new ArrayList<>();
		}
		
		changeListeners.add(listener);
	}

	/**
	 * @see de.ims.icarus.util.data.DataList#removeChangeListener(javax.swing.event.ChangeListener)
	 */
	@Override
	public void removeChangeListener(ChangeListener listener) {
		if(changeListeners==null) {
			return;
		}
		
		changeListeners.remove(listener);
	}

	/**
	 * @see de.ims.icarus.language.SentenceDataList#supportsType(de.ims.icarus.language.DataType)
	 */
	@Override
	public boolean supportsType(DataType type) {
		Treebank treebank = getTreebank();
		return treebank==null ? false : treebank.supportsType(type);
	}

	/**
	 * @see de.ims.icarus.language.SentenceDataList#get(int, de.ims.icarus.language.DataType)
	 */
	@Override
	public SentenceData get(int index, DataType type) {
		Treebank treebank = getTreebank();
		return treebank==null ? null : treebank.get(index, type);
	}

	/**
	 * @see de.ims.icarus.language.SentenceDataList#get(int, de.ims.icarus.language.DataType, de.ims.icarus.language.AvailabilityObserver)
	 */
	@Override
	public SentenceData get(int index, DataType type,
			AvailabilityObserver observer) {
		Treebank treebank = getTreebank();
		return treebank==null ? null : treebank.get(index, type, observer);
	}

	/**
	 * @see de.ims.icarus.util.NamedObject#getName()
	 */
	@Override
	public String getName() {
		Treebank treebank = getTreebank();
		return treebank==null ? null : treebank.getName();
	}

	/**
	 * @see de.ims.icarus.io.Loadable#isLoaded()
	 */
	@Override
	public boolean isLoaded() {
		Treebank treebank = getTreebank();
		return treebank==null ? false : treebank.isLoaded();
	}

	/**
	 * @see de.ims.icarus.io.Loadable#isLoading()
	 */
	@Override
	public boolean isLoading() {
		Treebank treebank = getTreebank();
		return treebank==null ? false : treebank.isLoading();
	}

	/**
	 * @see de.ims.icarus.io.Loadable#load()
	 */
	@Override
	public void load() throws Exception {
		Treebank treebank = getTreebank();
		if(treebank!=null) {
			treebank.load();
		}
	}

	/**
	 * @see de.ims.icarus.io.Loadable#free()
	 */
	@Override
	public void free() {
		Treebank treebank = getTreebank();
		if(treebank!=null) {
			treebank.free();
		}
	}
	
	@Override
	public String toString() {
		Treebank treebank = getTreebank();
		return treebank==null ? "<empty>" : treebank.getName(); //$NON-NLS-1$
	}
	
	private void fireChangeEvent() {
		if(changeListeners==null || changeListeners.isEmpty()) {
			return;
		}
		ChangeEvent event = new ChangeEvent(this);
		Object[] listeners = changeListeners.toArray();
		for(Object listener : listeners) {
			((ChangeListener)listener).stateChanged(event);
		}
	}
	
	/**
	 * 
	 * @author Markus Gärtner
	 * @version $Id: TreebankListDelegate.java 189 2013-12-04 15:37:42Z mcgaerty $
	 *
	 */
	private static class OwnedChangeListener implements ChangeListener {
		
		private final Reference<TreebankListDelegate> ref;
		
		OwnedChangeListener(TreebankListDelegate delegate) {
			if(delegate==null)
				throw new NullPointerException("Invalid delegate"); //$NON-NLS-1$
			
			ref = new WeakReference<>(delegate);
		}
		
		void destroy() {
			ref.clear();
		}

		/**
		 * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
		 */
		@Override
		public void stateChanged(ChangeEvent e) {
			TreebankListDelegate delegate = ref.get();
			if(delegate==null) {
				Treebank target = (Treebank)e.getSource();
				target.removeChangeListener(this);
			} else {
				delegate.fireChangeEvent();
			}
		}		
	}
}
